
import java.awt.image.BufferedImage;
import java.io.File;
import java.util.ArrayList;

import javax.imageio.ImageIO;

import javafx.application.Application;
import javafx.application.Platform;
import javafx.embed.swing.SwingFXUtils;
import javafx.event.ActionEvent;
import javafx.event.EventHandler;
import javafx.geometry.Rectangle2D;
import javafx.scene.Node;
import javafx.scene.Scene;
import javafx.scene.canvas.Canvas;
import javafx.scene.canvas.GraphicsContext;
import javafx.scene.control.Alert;
import javafx.scene.control.CheckBox;
import javafx.scene.control.ColorPicker;
import javafx.scene.control.ComboBox;
import javafx.scene.control.Label;
import javafx.scene.control.Menu;
import javafx.scene.control.MenuBar;
import javafx.scene.control.MenuItem;
import javafx.scene.control.SeparatorMenuItem;
import javafx.scene.control.TextField;
import javafx.scene.image.Image;
import javafx.scene.image.ImageView;
import javafx.scene.layout.BorderPane;
import javafx.scene.layout.HBox;
import javafx.scene.layout.StackPane;
import javafx.scene.layout.VBox;
import javafx.scene.paint.Color;
import javafx.scene.paint.ImagePattern;
import javafx.scene.paint.Paint;
import javafx.scene.shape.Ellipse;
import javafx.scene.shape.Rectangle;
import javafx.scene.text.Font;
import javafx.scene.text.FontPosture;
import javafx.scene.text.FontWeight;
import javafx.scene.text.Text;
import javafx.stage.FileChooser;
import javafx.stage.Screen;
import javafx.stage.Stage;

/**
 * Lets the user design images that can serve as "eCards".  The image can have
 * a colored or textured background, and it can contain images, shapes, and text
 * items.  The user can drag the items around in the image.  When a text item
 * or shape is added to the image, its properties are determined by the settings
 * of several controls in a toolbar at the bottom of the window.
 */
public class ECardDesigner extends Application {
	
	private ArrayList<Node> items;  // The items on the card (not including background).

	private Canvas background;  // Canvas where the background of the card is drawn.
	private StackPane content;  // Holds the canvas an all the items on the card.
	private Paint backgroundPaint; // Background color or texture for panel.
	private Stage window;  // The application window (needed for file selection dialog).
	
	private MenuItem removeItemCommand; // Menu item for removing the front item; this
										// is disabled when there are no items on the card.
	
	private TextField textInput;  // User input for a new text item to be added to the card.
	private CheckBox italicCheck; // Determines whether a new text item is italic.
	private CheckBox boldCheck;   // Determines whether a new text item is bold.
	private ComboBox<String> textSize;   // Determines the font size for a new text item.
	private ColorPicker textColor; // Determines the color of a new text item.
	
	private TextField widthInput;      // User input for the width of a new shape.
	private TextField heightInput;     // User inout for the height of a new shape.
	private ColorPicker shapeColor;    // Determines the color of a new shape.
	
	private final Color[] backgroundColors = { // Standard background colors.
	     Color.WHITE, Color.BLACK, Color.rgb(255,255,180), Color.PINK, Color.LIGHTGRAY,
	     Color.rgb(180,180,255), Color.rgb(130,225,130), Color.rgb(240,240,220)
	};
	
	private final String[] backgroundColorNames = { // Names corresponding to backgroundColors.
		 "White", "Black", "Yellow", "Pink",
		 "Gray", "Blue", "Green", "Beige"
	};
	
	private final String[] textureFileNames = {  // File names for standard background textures.
			 "beige_paper.jpg", "blue_paper.jpg", "bolts.gif", "brushed-metal.jpg", "celtic-knot.jpg",
			 "cinderblock.jpg", "gray_mosaic.jpg", "marble.jpg", "pink_marble.gif", 
			 "pink_roses.jpg", "sky.jpg", "wood_floor.jpg"
		};
	
	private final String[] imageFileNames = {  // File names for available images for ImageItems.
			 "canyon.jpg", "coxe.jpg", "earth.png", "earthrise.jpg", "galaxy.jpg", "hws.jpg", "mandelbrot1.jpg", 
			 "mandelbrot2.jpg", "mandelbrot3.jpg",  "quad.jpg", "seneca.jpg", "space_station.jpg"
		};
	
	
	public static void main(String[] args) { // Required to run a JavaFX program.
		launch();
	}
	
	/** 
	 * Sets up the application window. 
	 */
	public void start(Stage stage) {
		window = stage;
		items = new ArrayList<>();
		backgroundPaint = Color.WHITE;
		Rectangle2D screenSize = Screen.getPrimary().getBounds();
		double width = 1000;  // Width and height might be adjusted to make sure
		double height = 600;  //    that the window actually fits on the screen.
		if (width > screenSize.getWidth() - 80) {// width is too big for screen
			width = screenSize.getWidth() - 80;
		}
		if (height > screenSize.getHeight() - 120) {// height is too big for screen
			height = screenSize.getHeight() - 120;
		}
		background = new Canvas(width,height);
		repaintBackground();
		content = new StackPane(background);
		BorderPane root = new BorderPane();
		root.setCenter(content);
		root.setBottom(getToolbar());
		root.setTop(getMenuBar());
		stage.setScene(new Scene(root));
		stage.setResizable(false);
		stage.setTitle("eCard Designer");
		stage.show();
		
		// TODO:  DELETE THE FOLLOWING CODE FROM THIS METHOD BEFORE TURNING IN !!!!!
		// This gives example of making draggable items of the three types:  an ImageView,
		// a Text, and a Shape. In this case the Shape is a Rectangle, which is a
		// subclass of shape.
		
		ImageView img = new ImageView(new Image("images/earthrise.jpg"));
		makeDraggable(img);
		content.getChildren().add(img);
		
		Text text = new Text("Hello World!\nGlad to meet you!");
		Font font = Font.font("Times New Roman",FontWeight.BOLD,FontPosture.ITALIC,60);
		text.setFont(font);
		text.setFill(Color.CADETBLUE); 
		text.setStroke(Color.BLACK);
		makeDraggable(text);
		content.getChildren().add(text);
		text.relocate(300,350); // Without this, the first line, 
		                        // will be off the top of the window. 
		
		Rectangle shape = new Rectangle(300,200);
		shape.setFill(Color.color(0.7,0,0.5));
		shape.setStroke(Color.BLACK);
		shape.setArcWidth(36);   // Make the corners rounded.
		shape.setArcHeight(36);
		shape.setStrokeWidth(2);
		makeDraggable(shape);
		content.getChildren().add(shape);
		shape.relocate(500,20);
		
	} // end start();

	
	//--------------- Toolbar --------------------------------------------------------
	
	/**
	 * Create the toolbar for the bottom of the window, consisting of two boxes
	 * of controls, one to control text properties and one to control shape
	 * properties.
	 */
	private VBox getToolbar() {
		textInput = new TextField();
		textInput.setPrefColumnCount(25);
		italicCheck = new CheckBox("Italic");
		boldCheck = new CheckBox("Bold");
		textSize = new ComboBox<>();
		textSize.getItems().addAll("16", "20", "24", "30", "36", "48", "60", "72");
		textSize.getSelectionModel().select(2);
		textColor = new ColorPicker(Color.BLACK);
		textColor.getStyleClass().add("button");
		HBox textControls = new HBox(10,new Label("Text ="), textInput, new Label("size:"),
				textSize, italicCheck, boldCheck, textColor);
		textControls.setStyle("-fx-padding:8px; -fx-background-color:#EEE; -fx-border-color:black");
		widthInput = new TextField("200");
		widthInput.setPrefColumnCount(5);
		heightInput = new TextField("150");
		heightInput.setPrefColumnCount(5);
		shapeColor = new ColorPicker(Color.RED);
		shapeColor.getStyleClass().add("button");
		HBox shapeControls = new HBox(10,new Label("Shape size = "), widthInput, new Label(" by "),
				heightInput, new Label("Fill:"), shapeColor);
		shapeControls.setStyle("-fx-padding:8px; -fx-background-color:#EEE; "
				+ "-fx-border-color:black; -fx-border-width: 0 1px 1px 1px");
		return new VBox(textControls,shapeControls);
	}
		
	
	//--------------- Menubar and menu handling ---------------------------------------
	
	/**
	 * Create and return a menu bar containing menus for this application.
	 */
	private MenuBar getMenuBar() {
		MenuBar menubar = new MenuBar();
		
		// The following event handler will be used for every menu item.
		// It responds to a menu item by calling the method handleMenu(text),
		// where text is the text that appears in the menu item.  All menu
		// items will be implemented in that method.  This is not really
		// good style, but at least it keeps all the menu-handling code
		// in one place.
		
		EventHandler<ActionEvent> listener = 
				e -> handleMenu(((MenuItem)e.getTarget()).getText());
				
		Menu fileMenu = new Menu("File");
		menubar.getMenus().add(fileMenu);
		MenuItem saveImage = new MenuItem("Save Image...");
		saveImage.setOnAction(listener);
		fileMenu.getItems().add(saveImage);
		MenuItem quit = new MenuItem("Quit");
		quit.setOnAction(listener);
		fileMenu.getItems().add(quit);

		Menu backgroundMenu = new Menu("Background");
		menubar.getMenus().add(backgroundMenu);
		for (String cmd : backgroundColorNames) { // add color backgrounds
			MenuItem item = new MenuItem(cmd);
			item.setOnAction(listener);
			backgroundMenu.getItems().add(item);
		}
		MenuItem customBG = new MenuItem("Custom Color...");
		customBG.setOnAction(listener);
		backgroundMenu.getItems().add(customBG);
		backgroundMenu.getItems().add(new SeparatorMenuItem());
		for (String cmd : textureFileNames) { // add texture backgrounds
			MenuItem item = new MenuItem(cmd);
			item.setOnAction(listener);
			backgroundMenu.getItems().add(item);
		}
				
		Menu editMenu = new Menu("Edit");
		menubar.getMenus().add(editMenu);
		removeItemCommand = new MenuItem("Remove Front Item");
		removeItemCommand.setOnAction(listener);
		editMenu.getItems().add(removeItemCommand);
		
		removeItemCommand.setDisable(true);  // Must be enabled when an item is added.
		
		Menu addMenu = new Menu("Add Item");
		menubar.getMenus().add(addMenu);
		String[] addCommands = new String[] {"Add Text", "Add Rect", "Add Oval", "Add RoundRect"};
		for (String cmd : addCommands ) {
			MenuItem edititem = new MenuItem(cmd);
			edititem.setOnAction(listener);
			addMenu.getItems().add(edititem);
		}
		addMenu.getItems().add(new SeparatorMenuItem());
		for (String filename : imageFileNames) {
			MenuItem addimage = new MenuItem(filename);
			addimage.setOnAction(listener);
			addMenu.getItems().add(addimage);
		}
		
		return menubar;
	}
	
	/**
	 * This class defines the listener that is used to respond to ALL menu commands.
	 * The parameter, command, is the text that appears in the menu command.
	 */
	private void handleMenu(String command) {
		if (command.equals("Save Image...")) {
			doSaveImage();
			return;
		}
		if (command.equals("Quit")) {
			Platform.exit();
			return;
		}
		for (int i = 0; i < backgroundColorNames.length; i++) {
			if (command.equals(backgroundColorNames[i])) {
				backgroundPaint = backgroundColors[i];
				repaintBackground();
				return;
			}
		}
		for (String filename : textureFileNames) {
			if (command.equals(filename)) {
				Image texture = new Image("textures/" + filename);
				backgroundPaint = new ImagePattern(texture, 0, 0, 
						texture.getWidth(), texture.getHeight(), false);
				repaintBackground();
				return;
			}
		}
		if (command.equals("Custom Color...")) {
			Color c = SimpleDialogs.colorChooser(Color.RED, "Select Background Color");
			if (c != null) {
				backgroundPaint = c;
				repaintBackground();
			}
			return;
		}
	}
	
	//---------------------------------------------------------------------

	/**
	 * Fill the canvas with the current background paint.  This should
	 * be called whenever the background is changed.
	 */
	private void repaintBackground() {
		GraphicsContext g = background.getGraphicsContext2D();
		g.setFill(backgroundPaint);
		g.fillRect(0,0,background.getWidth(),background.getHeight());
	}
	


    /**
     * Saves the user's e-card as an image file in PNG format.
     */
    private void doSaveImage() {
        FileChooser fileDialog = new FileChooser(); 
        fileDialog.setInitialFileName("imagefile.png");
        fileDialog.setInitialDirectory(
                         new File( System.getProperty("user.home") ) );
        fileDialog.setTitle("Select File to Save. Name MUST end with .png!");
        File selectedFile = fileDialog.showSaveDialog(window);
        if ( selectedFile == null )
            return;  // User did not select a file.
        try {
            Image contentImage = content.snapshot(null,null);
            BufferedImage image = SwingFXUtils.fromFXImage(contentImage,null);
            String filename = selectedFile.getName().toLowerCase();
            if ( ! filename.endsWith(".png")) {
                throw new Exception("The file name must end with \".png\".");
            }
            boolean hasFormat = ImageIO.write(image,"PNG",selectedFile);
            if ( ! hasFormat ) { // (this should never happen)
                throw new Exception( "PNG format not available.");
            }
        }
        catch (Exception e) {
            Alert errorAlert = new Alert(Alert.AlertType.ERROR,
                   "Sorry, an error occurred while\ntrying to save the image:\n"
                         + e.getMessage());
            errorAlert.showAndWait();
        }    
    }
    

    /**
     * Call this make it possible for the user to drag an item within the
     * StackPane that contains it; otherwise it will just sit in the center.
     * The node must still be added to the children list of the StackPane.
     *    (It was difficult to arrange for the items that are added to the e-card
     * to be draggable by the user.  First of all, they have to be "unmanaged"
     * so that the StackPane does not force them to the center of the pain.
     * The position of an unmanaged node can be set using the relocate() method.
     * Furthermore, the node needs mouse event handlers that respond when the
     * user drags the node.  Since the motion is really in terms of the
     * StackPane's coordinate system, this involves some tricky manipulation
     * of the coordinates.  (This might not work in all circumstances.))
     */
    private static void makeDraggable(Node node) {
		final double[] position = new double[2]; // Save previous mouse coords.
		   // Note that the coordinates can't be saved in separate double
		   // variables, since a local variable that is used in a lambda
		   // expression must be "final" or effectively final.  That is,
		   // its value can't ever be changed from its initial value.
		node.setManaged(false);  // So the StackPane won't try to set node's position.
		node.setOnMousePressed( e -> {
			position[0] = node.getBoundsInParent().getMinX() + e.getX();
			position[1] = node.getBoundsInParent().getMinY() + e.getY();
		});
		node.setOnMouseDragged( e -> {
			double x = node.getBoundsInParent().getMinX() + e.getX();
			double y = node.getBoundsInParent().getMinY() + e.getY();
			double dx = x - position[0];
			double dy = y - position[1];
			node.setLayoutX(node.getLayoutX() + dx);
			node.setLayoutY(node.getLayoutY() + dy);
			position[0] = x;
			position[1] = y;
		});
	}

}
